#! /usr/bin/env python
#
# Author: Roberto Cuccu - RSS Team - 2018

import os
import re
import glob
import numpy as np
import io
import sys


class L8_Product:

    def __init__(self, product_path):

        # Check if path exist
        if not os.path.exists(product_path):
            raise (IOError('Product path does not exist'))

        # Get L8 product structure
        try:
            self.path = product_path

            file_to_match = 'L*_B1.TIF'
            band_1 = glob.glob(os.path.join(product_path, file_to_match))[0]
            self.band_1 = band_1
            file_to_match = 'L*_B2.TIF'
            band_2 = glob.glob(os.path.join(product_path, file_to_match))[0]
            self.band_2 = band_2
            file_to_match = 'L*_B3.TIF'
            band_3 = glob.glob(os.path.join(product_path, file_to_match))[0]
            self.band_3 = band_3
            file_to_match = 'L*_B4.TIF'
            band_4 = glob.glob(os.path.join(product_path, file_to_match))[0]
            self.band_4 = band_4
            file_to_match = 'L*_B5.TIF'
            band_5 = glob.glob(os.path.join(product_path, file_to_match))[0]
            self.band_5 = band_5
            file_to_match = 'L*_B6.TIF'
            band_6 = glob.glob(os.path.join(product_path, file_to_match))[0]
            self.band_6 = band_6
            file_to_match = 'L*_B7.TIF'
            band_7 = glob.glob(os.path.join(product_path, file_to_match))[0]
            self.band_7 = band_7
            file_to_match = 'L*_B10.TIF'
            band_10 = glob.glob(os.path.join(product_path, file_to_match))[0]
            self.band_10 = band_10
            file_to_match = 'L*_B11.TIF'
            band_11 = glob.glob(os.path.join(product_path, file_to_match))[0]
            self.band_11 = band_11

            self.bands = {
                "b1": band_1,
                "b2": band_2,
                "b3": band_3,
                "b4": band_4,
                "b5": band_5,
                "b6": band_6,
                "b7": band_7,
                "b10": band_10,
                "b11": band_11
            }

        except IndexError:
            os.sys.exit('Error - Not possible to retrieve L8 product structure - Procedure aborted')


class L8_MTL:

    def __init__(self, MTL_path):

        # Get MTL text file
        MTL_string_match = 'L*_MTL.txt'
        MTL_file_search = glob.glob(os.path.join(MTL_path, MTL_string_match))
        MTL_file = MTL_file_search[0]

        if not os.path.exists(MTL_file):
            raise (IOError('MTL file not found'))

        try:
            self.MTL_path = MTL_path
            self.MTL_file = MTL_file

            # Read MTL file
            MTL_file_open = io.open(MTL_file, 'rU')
            MTL_content = MTL_file_open.read()
            MTL_file_open.close()

            # Get metadata
            # REFLECTANCE MULT BANDS
            string_to_search = 'REFLECTANCE_MULT_BAND_1 =.*'
            reflectance_mult_band_1 = get_mtl_value(MTL_content, string_to_search)
            self.reflectance_mult_band_1 = reflectance_mult_band_1

            string_to_search = 'REFLECTANCE_MULT_BAND_2 =.*'
            reflectance_mult_band_2 = get_mtl_value(MTL_content, string_to_search)
            self.reflectance_mult_band_2 = reflectance_mult_band_2

            string_to_search = 'REFLECTANCE_MULT_BAND_3 =.*'
            reflectance_mult_band_3 = get_mtl_value(MTL_content, string_to_search)
            self.reflectance_mult_band_3 = reflectance_mult_band_3

            string_to_search = 'REFLECTANCE_MULT_BAND_4 =.*'
            reflectance_mult_band_4 = get_mtl_value(MTL_content, string_to_search)
            self.reflectance_mult_band_4 = reflectance_mult_band_4

            string_to_search = 'REFLECTANCE_MULT_BAND_5 =.*'
            reflectance_mult_band_5 = get_mtl_value(MTL_content, string_to_search)
            self.reflectance_mult_band_5 = reflectance_mult_band_5

            string_to_search = 'REFLECTANCE_MULT_BAND_6 =.*'
            reflectance_mult_band_6 = get_mtl_value(MTL_content, string_to_search)
            self.reflectance_mult_band_6 = reflectance_mult_band_6

            string_to_search = 'REFLECTANCE_MULT_BAND_7 =.*'
            reflectance_mult_band_7 = get_mtl_value(MTL_content, string_to_search)
            self.reflectance_mult_band_7 = reflectance_mult_band_7

            string_to_search = 'RADIANCE_MULT_BAND_10 =.*'
            reflectance_mult_band_10 = get_mtl_value(MTL_content, string_to_search)
            self.reflectance_mult_band_10 = reflectance_mult_band_10

            string_to_search = 'RADIANCE_MULT_BAND_11 =.*'
            reflectance_mult_band_11 = get_mtl_value(MTL_content, string_to_search)
            self.reflectance_mult_band_11 = reflectance_mult_band_11

            string_to_search = 'K1_CONSTANT_BAND_10 =.*'
            reflectance_k1_band_10 = get_mtl_value(MTL_content, string_to_search)
            self.reflectance_k1_band_10 = reflectance_k1_band_10

            string_to_search = 'K1_CONSTANT_BAND_11 =.*'
            reflectance_k1_band_11 = get_mtl_value(MTL_content, string_to_search)
            self.reflectance_k1_band_11 = reflectance_k1_band_11

            self.reflectance_mult_factors = {
                "REFLECTANCE_MULT_BAND_1": reflectance_mult_band_1,
                "REFLECTANCE_MULT_BAND_2": reflectance_mult_band_2,
                "REFLECTANCE_MULT_BAND_3": reflectance_mult_band_3,
                "REFLECTANCE_MULT_BAND_4": reflectance_mult_band_4,
                "REFLECTANCE_MULT_BAND_5": reflectance_mult_band_5,
                "REFLECTANCE_MULT_BAND_6": reflectance_mult_band_6,
                "REFLECTANCE_MULT_BAND_7": reflectance_mult_band_7,
                "RADIANCE_MULT_BAND_10": reflectance_mult_band_10,
                "RADIANCE_MULT_BAND_11": reflectance_mult_band_11,
                "K1_CONSTANT_BAND_10": reflectance_k1_band_10,
                "K1_CONSTANT_BAND_11": reflectance_k1_band_11
            }

            # REFLECTANCE ADD BANDS
            string_to_search = 'REFLECTANCE_ADD_BAND_1 =.*'
            reflectance_add_band_1 = get_mtl_value(MTL_content, string_to_search)
            self.reflectance_add_band_1 = reflectance_add_band_1

            string_to_search = 'REFLECTANCE_ADD_BAND_2 =.*'
            reflectance_add_band_2 = get_mtl_value(MTL_content, string_to_search)
            self.reflectance_add_band_2 = reflectance_add_band_2

            string_to_search = 'REFLECTANCE_ADD_BAND_3 =.*'
            reflectance_add_band_3 = get_mtl_value(MTL_content, string_to_search)
            self.reflectance_add_band_3 = reflectance_add_band_3

            string_to_search = 'REFLECTANCE_ADD_BAND_4 =.*'
            reflectance_add_band_4 = get_mtl_value(MTL_content, string_to_search)
            self.reflectance_add_band_4 = reflectance_add_band_4

            string_to_search = 'REFLECTANCE_ADD_BAND_5 =.*'
            reflectance_add_band_5 = get_mtl_value(MTL_content, string_to_search)
            self.reflectance_add_band_5 = reflectance_add_band_5

            string_to_search = 'REFLECTANCE_ADD_BAND_6 =.*'
            reflectance_add_band_6 = get_mtl_value(MTL_content, string_to_search)
            self.reflectance_add_band_6 = reflectance_add_band_6

            string_to_search = 'REFLECTANCE_ADD_BAND_7 =.*'
            reflectance_add_band_7 = get_mtl_value(MTL_content, string_to_search)
            self.reflectance_add_band_7 = reflectance_add_band_7

            string_to_search = 'RADIANCE_ADD_BAND_10 =.*'
            reflectance_add_band_10 = get_mtl_value(MTL_content, string_to_search)
            self.reflectance_add_band_10 = reflectance_add_band_10

            string_to_search = 'RADIANCE_ADD_BAND_11 =.*'
            reflectance_add_band_11 = get_mtl_value(MTL_content, string_to_search)
            self.reflectance_add_band_11 = reflectance_add_band_11

            string_to_search = 'K2_CONSTANT_BAND_10 =.*'
            reflectance_k2_band_10 = get_mtl_value(MTL_content, string_to_search)
            self.reflectance_k2_band_10 = reflectance_k2_band_10

            string_to_search = 'K2_CONSTANT_BAND_11 =.*'
            reflectance_k2_band_11 = get_mtl_value(MTL_content, string_to_search)
            self.reflectance_k2_band_11 = reflectance_k2_band_11

            self.reflectance_add_factors = {
                "REFLECTANCE_ADD_BAND_1": reflectance_add_band_1,
                "REFLECTANCE_ADD_BAND_2": reflectance_add_band_2,
                "REFLECTANCE_ADD_BAND_3": reflectance_add_band_3,
                "REFLECTANCE_ADD_BAND_4": reflectance_add_band_4,
                "REFLECTANCE_ADD_BAND_5": reflectance_add_band_5,
                "REFLECTANCE_ADD_BAND_6": reflectance_add_band_6,
                "REFLECTANCE_ADD_BAND_7": reflectance_add_band_7,
                "RADIANCE_ADD_BAND_10": reflectance_add_band_10,
                "RADIANCE_ADD_BAND_11": reflectance_add_band_11,
                "K2_CONSTANT_BAND_10": reflectance_k2_band_10,
                "K2_CONSTANT_BAND_11": reflectance_k2_band_11
            }

            # SUN ELEVATION and ZENITH angles
            string_to_search = 'SUN_ELEVATION =.*'
            sun_elevation = get_mtl_value(MTL_content, string_to_search)
            sun_zenith = 90.0 - np.float(sun_elevation)
            self.sun_elevation_angle = sun_elevation
            self.sun_zenith_angle = sun_zenith

        except IndexError:
            os.sys.exit('Error - Not possible to retriev MTL metadata - Procedure aborted')


def get_mtl_value(MTL, pattern2match):
    check = re.compile(pattern2match)
    match = (check.findall(MTL))
    if match:
        mtl_value = (match[0].split('='))[1].replace('"', '').replace(' ', '')
    else:
        mtl_value = 'no match'
    return mtl_value




#! /usr/bin/env python
# 
# Author: Roberto Cuccu - RSS Team - 2018

from argparse import ArgumentParser
import os, sys, shutil
import numpy as np
import gdal
from L8_reflectance.L8_utils import L8_MTL, L8_Product


def process_band(myband, mult, add, se, outband, K1, K2, thermal):
    print
    myband + ' | ' + mult + ' | ' + add + ' | ' + se + ' | ' + outband

    # Open input band with GDAL
    ds = gdal.Open(myband)
    band = ds.GetRasterBand(1)
    arr = band.ReadAsArray()
    [cols, rows] = arr.shape

    # Calculate TOA reflectance band
    # ------------------------------------------------------------------
    # Reference: https://landsat.usgs.gov/using-usgs-landsat-8-product
    # ------------------------------------------------------------------
    if not thermal:
        rf = ((np.float32(mult) * arr.astype(np.float32)) + np.float32(add)) / np.sin(np.deg2rad(np.float32(se)))
        gdt = gdal.GDT_Float32
    else:
        toa_radiance = (np.float32(mult) * arr.astype(np.float32)) + np.float32(add)
        rf = np.float32(K2) / np.log(np.float32(K1)/toa_radiance + 1)
        gdt = gdal.GDT_Float32

    # Write Output TOA reflectanf band in GTiff Float32
    driver = gdal.GetDriverByName("GTiff")
    outdata = driver.Create(outband, rows, cols, 1, gdt)
    outdata.SetGeoTransform(ds.GetGeoTransform())  # Set same GEO transformation coeff as input
    outdata.SetProjection(ds.GetProjection())  # Set same projection as input
    outdata.GetRasterBand(1).WriteArray(rf)
    outdata.FlushCache()

    # Release variables
    outdata = None
    band = None
    ds = None


def main():
    # parse arguments
    parser = ArgumentParser()
    parser.add_argument('product', help="Input Landsat-8 product path")
    parser.add_argument('outproduct', help="Output L8 TOA reflectance product path")
    args = parser.parse_args()

    # Get L8 product and its metadata	
    L8_prod = L8_Product(args.product)
    L8_metadata = L8_MTL(args.product)

    # Check if the output product dir exist otherwise create it
    outprodpath = args.outproduct

    if not os.path.exists(outprodpath):
        os.makedirs(outprodpath)

    # Create L8 TOA output product folder
    input_prod = args.product
    base_input_prod = os.path.basename(input_prod)
    out_prod_toa = base_input_prod + '_TOA'
    L8_outprodpath = os.path.join(outprodpath, out_prod_toa)

    if os.path.exists(L8_outprodpath):
        print
        "Found existing directory {} . cleaning it and recreating it..".format(L8_outprodpath)
        shutil.rmtree(L8_outprodpath)
        os.makedirs(L8_outprodpath)
    else:
        os.makedirs(L8_outprodpath)

    # Retrieve sun elevation angle from metadata
    se = L8_metadata.sun_elevation_angle

    # L8 reflectance band indexes
    bi = [1, 2, 3, 4, 5, 6, 7, 10, 11]

    # Loop over bands
    for i in bi:
        # Retrieve input band location
        curr_band = L8_prod.bands["b{}".format(i)]

        # Determine output band location
        input_band_filename = os.path.basename(curr_band)
        output_band_filename = input_band_filename.replace(".TIF", "_TOA.TIF")
        output_band_path = os.path.join(L8_outprodpath, output_band_filename)

        # Retrieve rescaling factors
        if i<9:
            mult = L8_metadata.reflectance_mult_factors["REFLECTANCE_MULT_BAND_{}".format(i)]
            add = L8_metadata.reflectance_add_factors["REFLECTANCE_ADD_BAND_{}".format(i)]
            # Process band
            process_band(curr_band, mult, add, se, output_band_path, None, None, False)
        else:
            mult = L8_metadata.reflectance_mult_factors["RADIANCE_MULT_BAND_{}".format(i)]
            add = L8_metadata.reflectance_add_factors["RADIANCE_ADD_BAND_{}".format(i)]
            K1 = L8_metadata.reflectance_mult_factors["K1_CONSTANT_BAND_{}".format(i)]
            K2 = L8_metadata.reflectance_add_factors["K2_CONSTANT_BAND_{}".format(i)]
            # Process band
            process_band(curr_band, mult, add, se, output_band_path, K1, K2, True)

    sys.exit(0)


if __name__ == "__main__":
    main()




try:
    from setuptools import setup

    setup_kwargs = {'entry_points': {'console_scripts': ['L8_reflectance=L8_reflectance.L8_reflectance:main']}}
except ImportError:
    from distutils.core import setup

    setup_kwargs = {'scripts': ['bin/L8_reflectance']}

tag = '0.0.5'

setup(name='L8_reflectance',
      version=tag,
      py_modules=['L8_reflectance'],
      description='Convert Landsat 8 from DN to reflectance TOA',
      url='https://github.com/ESRIN-RSS/L8_reflectance',
      author='Roberto Cuccu',
      author_email='roberto.cuccu@esa.int',
      license='GPL',
      zip_safe=False,
      packages=['L8_reflectance'],
      install_requires=['numpy', 'gdal'],
      dependency_links=['https://www.conan.io/source/Gdal/2.1.3/osechet/stable'],
      **setup_kwargs)
